package POE::Session::SubVar;

use warnings;
use strict;
use vars qw($VERSION @ISA);
use POE::Preprocessor ( isa => 'POE::Session' );
use POSIX qw(ENOSYS);

require POE::Session;
push @ISA, 'POE::Session';
$VERSION = '0.01';

# duplicate constants from POE::Session (bad?)

sub SE_NAMESPACE    () { 0 }
sub SE_OPTIONS      () { 1 }
sub SE_STATES       () { 2 }

sub OPT_DEFAULT     () { 'default' }
sub OPT_TRACE       () { 'trace' }

sub EN_DEFAULT      () { '_default' }
sub EN_SIGNAL       () { '_signal' }


my @context;

sub import {
  my $package;
  for ( my $i=0; $package = [caller($i)]->[0]; $i++ ) {
    last unless $package->isa('POE::Session') 
  }
  no strict 'refs';
  *{ $package . '::session'     } = sub { $context[-1]->[0] };
  *{ $package . '::kernel'      } = sub { $context[-1]->[1] };
  *{ $package . '::heap'        } = sub { $context[-1]->[2] };
  *{ $package . '::state'       } = sub { $context[-1]->[3] };
  *{ $package . '::sender'      } = sub { $context[-1]->[4] };
  *{ $package . '::caller_file' } = sub { $context[-1]->[5] };
  *{ $package . '::caller_line' } = sub { $context[-1]->[6] };
}

sub _invoke_state {
  my ($self, $source_session, $state, $etc, $file, $line) = @_;

  # Trace the state invocation if tracing is enabled.

  if ($self->[SE_OPTIONS]->{+OPT_TRACE}) {
    warn( $POE::Kernel::poe_kernel->ID_session_to_id($self),
          " -> $state (from $file at $line)\n"
        );
  }

  # The desired destination state doesn't exist in this session.
  # Attempt to redirect the state transition to _default.

  unless (exists $self->[SE_STATES]->{$state}) {

    # There's no _default either; redirection's not happening today.
    # Drop the state transition event on the floor, and optionally
    # make some noise about it.

    unless (exists $self->[SE_STATES]->{+EN_DEFAULT}) {
      $! = ENOSYS;
      if ($self->[SE_OPTIONS]->{+OPT_DEFAULT} and $state ne EN_SIGNAL) {
        warn( "a '$state' state was sent from $file at $line to session ",
              $POE::Kernel::poe_kernel->ID_session_to_id($self),
              ", but session ",
              $POE::Kernel::poe_kernel->ID_session_to_id($self),
              " has neither that state nor a _default state to handle it\n"
            );
      }
      return undef;
    }

    # If we get this far, then there's a _default state to redirect
    # the transition to.  Trace the redirection.

    if ($self->[SE_OPTIONS]->{+OPT_TRACE}) {
      warn( $POE::Kernel::poe_kernel->ID_session_to_id($self),
            " -> $state redirected to _default\n"
          );
    }

    # Transmogrify the original state transition into a corresponding
    # _default invocation.

    $etc   = [ $state, $etc ];
    $state = EN_DEFAULT;
  }
 
  # If we get this far, then the state can be invoked.  Setup context
  # so parameter shortcuts may be used, and invoke it. (Per Kernel.pm
  # scalar return values are assumed)

  my $return;
  push @context, [
    $self,                          # session
    $POE::Kernel::poe_kernel,       # kernel
    $self->[SE_NAMESPACE],          # heap
    $state,                         # state
    $source_session,                # sender
    $file,                          # caller file name
    $line,                          # caller file line
  ];
 
  if (ref($self->[SE_STATES]->{$state}) eq 'CODE') {
    # Inline states are invoked this way.
    $return = $self->[SE_STATES]->{$state}->(@$etc);
  }
  else {
    # Package and object states are invoked this way.
    my ($class_or_object, $method) = @{$self->[SE_STATES]->{$state}};
    $return = $class_or_object->$method(@$etc);
  }

  pop @context;
  $return;
}
 
1;
__END__

=head1 NAME

POE::Session::SubVar;

=head1 VERSION

This document describes version 0.01 of POE::Session::SubVar,
released November 12, 2002.

=head1 SYNOPSIS

  use POE::Kernel;
  use POE::Session::SubVar;

  POE::Session::SubVar->create(
      inline_states => {
          _start => sub {
              warn "\@_: @_";
              kernel->yield('count' => 1);
          },
          count => sub {
              my ($count) = @_;
              print "counted to $count\n";
              kernel->yield(count => ++$count)  if $count < 10;
          },
          _stop => sub { warn "done" },
      },
      args => [qw(foo bar baz)],
  );

  $poe_kernel->run();

=head1 DESCRIPTION

POE::Session::SubVar is derived from POE::Session. Its chief purpose is
to provide subroutines to replace the passing of special variables when
event handlers are invoked.

The following conversion table may be useful:

  $_[SESSION]     ->  session() 
  $_[KERNEL]      ->  kernel() 
  $_[HEAP]        ->  heap() 
  $_[STATE]       ->  state() 
  $_[SENDER]      ->  sender() 
  $_[CALLER_FILE] ->  caller_file() 
  $_[CALLER_LINE] ->  caller_line() 

There is no conversion for $_[OBJECT] and @_[ARG0..$#_], because
these should now be transparent. Or to use an incomplete example:

  POE::Session->new(
      inline_state => sub {
          my (@args) = @_;
          heap->{args} = \@args;
          kernel->yield('object_state');
      },
      $object => ['object_state'],
  );
  
  sub object_state {
      my ($self, @args) = @_;
      ...
  }

=head1 AUTHOR

C. Garrett Goebel (ggoebel@cpan.org)

=head1 BUGS

This module has received precious little testing. 
Bug reports and other feedback are most welcome.

=head1 COPYRIGHT

Copyright (c) 2002, C. Garrett Goebel. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the terms of the Perl Artistic License
  (see http://www.perl.com/perl/misc/Artistic.html)

